__all__ = ('PermissionOverwrite', )

from scarletio import RichAttributeErrorBaseType, include

from ...permission import Permission

from .fields.allow import parse_allow, put_allow_into, validate_allow
from .fields.deny import parse_deny, put_deny_into, validate_deny
from .fields.target import validate_target
from .fields.target_id import parse_target_id, put_target_id_into, validate_target_id
from .fields.target_type import parse_target_type, put_target_type_into, validate_target_type
from .preinstanced import PermissionOverwriteTargetType


create_partial_role_from_id = include('create_partial_role_from_id')
create_partial_user_from_id = include('create_partial_user_from_id')

Role = include('Role')
ClientUserBase = include('ClientUserBase')

PERMISSION_OVERWRITE_TYPE_ROLE = PermissionOverwriteTargetType.role
PERMISSION_OVERWRITE_TYPE_USER = PermissionOverwriteTargetType.user


class PermissionOverwrite(RichAttributeErrorBaseType):
    """
    Represents a permission overwrite of a guild channel.
    
    Attributes
    ----------
    allow : ``Permission``
        The allowed permissions by the overwrite.
    deny : ``Permission``
        The denied permission by the overwrite.
    target_id : `int`
        The permission overwrites target's identifier.
    target_type : ``PermissionOverwriteTargetType``
        The permission overwrite's target's type.
    """
    __slots__ = ('allow', 'deny', 'target_id', 'target_type')
    
    def __new__(cls, target_id, *, target_type = None, allow = None, deny = None):
        """
        Creates a new permission overwrite from the given parameters.
        
        Parameters
        ----------
        target_id : `int`, ``Role``, ``ClientUserBase``
            The permission overwrite's target.
        target_type : `None`, ``PermissionOverwriteTargetType`` = `None`, Optional (Keyword only)
            The permission overwrite's target's type. Required if `target_id` is given as a snowflake.
        allow : `None`, ``Permission``, `int` = `None`, Optional (Keyword only)
            The permission overwrite's allowed permission's value.
        deny : `None`, ``Permission``, `int` = `None`, Optional (Keyword only)
            The permission overwrite's denied permission's value.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        target_id_processed, target_type_processed = validate_target(target_id)
        
        if (target_type is None):
            if target_type_processed is PermissionOverwriteTargetType.unknown:
                raise ValueError(
                    f'`target_type` cannot be `None` if `target_id` is given as a snowflake. '
                    f'Got target_id = {target_id!r}.'
                )
        
        else:
            target_type_processed_from_target_type = validate_target_type(target_type)
            if target_type_processed is PermissionOverwriteTargetType.unknown:
                target_type_processed = target_type_processed_from_target_type
            else:
                if target_type_processed is not target_type_processed_from_target_type:
                    raise ValueError(
                        f'If `target_id` is given as an entity, then `target_type` should not be given, or should '
                        f'match the entity\'s type. Got target_id = {target_id!r}; target_type = {target_type!r}.'
                    )
                    
        
        allow = validate_allow(allow)
        deny = validate_deny(deny)
        
        self = object.__new__(cls)
        self.target_id = target_id_processed
        self.target_type = target_type_processed
        self.allow = allow
        self.deny = deny
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a permission overwrite from the given data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received permission overwrite data.
        
        Returns
        -------
        self : ``PermissionOverwrite``
        """
        self = object.__new__(cls)
        
        self.target_id = parse_target_id(data)
        self.target_type = parse_target_type(data)
        self.allow = parse_allow(data)
        self.deny = parse_deny(data)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the permission overwrite to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default fields should be included as well.
            
            > This parameter has no effect on any fields.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields (like id-s) should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        put_allow_into(self.allow, data, defaults)
        put_deny_into(self.deny, data, defaults)
        put_target_id_into(self.target_id, data, defaults, include_internals = include_internals)
        put_target_type_into(self.target_type, data, defaults)
        
        return data
    
    
    @property
    def target(self):
        """
        Returns the target entity of the overwrite.
        
        Returns
        -------
        target : ``ClientUserBase``, ``Role``, `None`
        """
        target_type = self.target_type
        target_id = self.target_id
        
        if target_type is PERMISSION_OVERWRITE_TYPE_ROLE:
            target = create_partial_role_from_id(target_id)
        
        elif target_type is PERMISSION_OVERWRITE_TYPE_USER:
            target = create_partial_user_from_id(target_id)
        
        else:
            target = None
        
        return target
    
    
    def __hash__(self):
        """Returns the permission overwrite's hash."""
        return self.target_id ^ self.allow ^ self.deny ^ (self.target_type.value << 16)
    
    
    def __repr__(self):
        """Returns the permission overwrite's representation."""
        return f'<{self.__class__.__name__} target={self.target!r}>'
    
    
    def keys(self):
        """
        Yields the permissions' names.
        
        This method is an iterable generator.
        
        Yields
        ------
        name : `str`
            Permissions' respective name.
        """
        yield from Permission.__keys__.keys()
    
    
    def values(self):
        """
        Yields position by position each permission's state. `+1` is yielded if the permission is enabled, `-1` if
        disabled and `0` if neither.
        
        This method is a generator.
        
        Yields
        ------
        state : `int`
            The permission's state.
            
            Can be one of the following:
            
            +-----------+-------+
            | Name      | Value |
            +===========+=======+
            | Enabled   | +1    |
            +-----------+-------+
            | None      | 0     |
            +-----------+-------+
            | Disabled  | -1    |
            +-----------+-------+
        """
        allow = self.allow
        deny = self.deny
        for index in Permission.__keys__.values():
            if (allow >> index) & 1:
                state = +1
            elif (deny >> index) & 1:
                state = -1
            else:
                state = 0
            
            yield state
    
    
    def items(self):
        """
        Yields the permission overwrite's items. What includes their name and their state. As state `+1` is yielded
        if the permission is enabled, `-1` if disabled and `0` if neither.
        
        This method is a generator.
        
        Yields
        -------
        name : `str`
            Permissions' respective name.
        state : `int`
            The permission's state.
            
            Can be one of the following:
            
            +-----------+-------+
            | Name      | Value |
            +===========+=======+
            | Enabled   | +1    |
            +-----------+-------+
            | None      | 0     |
            +-----------+-------+
            | Disabled  | -1    |
            +-----------+-------+
        """
        allow = self.allow
        deny = self.deny
        for key, index in Permission.__keys__.items():
            if (allow >> index) & 1:
                state = +1
            elif (deny >> index) & 1:
                state = -1
            else:
                state = 0
            
            yield key, state
    
    
    def __getitem__(self, key):
        """Returns the permission's state for the given permission name."""
        index = Permission.__keys__[key]
        if (self.allow >> index) & 1:
            state = +1
        elif (self.deny >> index) & 1:
            state = -1
        else:
            state = 0
        
        return state
    
    
    def __lt__(self, other):
        """Returns whether is this permission overwrite is at lower position in ordering than the order."""
        if type(self) is not type(other):
            return NotImplemented
        
        target_type = self.target_type
        if target_type is PERMISSION_OVERWRITE_TYPE_USER:
            self_target_type_sort_value = 0
        
        elif target_type is PERMISSION_OVERWRITE_TYPE_ROLE:
            self_target_type_sort_value = 1
        
        else:
            self_target_type_sort_value = 2
        

        target_type = other.target_type
        if target_type is PERMISSION_OVERWRITE_TYPE_USER:
            other_target_type_sort_value = 0
        
        elif target_type is PERMISSION_OVERWRITE_TYPE_ROLE:
            other_target_type_sort_value = 1
        
        else:
            other_target_type_sort_value = 2
        
        
        if self_target_type_sort_value < other_target_type_sort_value:
            return True
        
        if self_target_type_sort_value == other_target_type_sort_value:
            if self.target_id < other.target_id:
                return True
            
            return False
        
        return False
    
    
    def __eq__(self, other):
        """Returns whether is this permission overwrite is same as the other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # allow
        if self.allow != other.allow:
            return False
        
        # deny
        if self.deny != other.deny:
            return False
        
        # target_id
        if self.target_id != other.target_id:
            return False
        
        # target_type
        if self.target_type is not other.target_type:
            return False
        
        return True
    
    
    def __gt__(self, other):
        """Returns whether is this permission overwrite is at greater position in ordering than the order."""
        if type(self) is not type(other):
            return NotImplemented
        
        target_type = self.target_type
        if target_type is PERMISSION_OVERWRITE_TYPE_USER:
            self_target_type_sort_value = 0
        
        elif target_type is PERMISSION_OVERWRITE_TYPE_ROLE:
            self_target_type_sort_value = 1
        
        else:
            self_target_type_sort_value = 2
        

        target_type = other.target_type
        if target_type is PERMISSION_OVERWRITE_TYPE_USER:
            other_target_type_sort_value = 0
        
        elif target_type is PERMISSION_OVERWRITE_TYPE_ROLE:
            other_target_type_sort_value = 1
        
        else:
            other_target_type_sort_value = 2
        
        
        if self_target_type_sort_value > other_target_type_sort_value:
            return True
        
        if self_target_type_sort_value == other_target_type_sort_value:
            if self.target_id > other.target_id:
                return True
            
            return False
        
        return False
    
    
    def copy(self):
        """
        Copies the permission overwrite.
        
        Returns
        -------
        new : `instance<cls<self>>`
        """
        new = object.__new__(type(self))
        new.allow = self.allow
        new.deny = self.deny
        new.target_id = self.target_id
        new.target_type = self.target_type
        return new
    
    
    def copy_with(self, **keyword_parameters):
        """
        Copies the permission overwrite modifying the given values.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            The specified fields to change.
        """
        allow = self.allow
        deny = self.deny
        target_id = self.target_id
        target_type = self.target_type
        
        # allow
        try:
            allow = keyword_parameters.pop('allow')
        except KeyError:
            pass
        else:
            allow = validate_allow(allow)
        
        # deny
        try:
            deny = keyword_parameters.pop('deny')
        except KeyError:
            pass
        else:
            deny = validate_deny(deny)
        
        # [EXTRA] target 
        try:
            target = keyword_parameters.pop('target')
        except KeyError:
            pass
        else:
            target_id, potential_target_type = validate_target(target)
            if (potential_target_type is not PermissionOverwriteTargetType.unknown):
                target_type = potential_target_type
        
        # target_id
        try:
            target_id = keyword_parameters.pop('target_id')
        except KeyError:
            pass
        else:
            target_id = validate_target_id(target_id)
        
        # target_type
        try:
            target_type = keyword_parameters.pop('target_type')
        except KeyError:
            pass
        else:
            target_type = validate_target_type(target_type)
        
        
        new = object.__new__(type(self))
        new.allow = allow
        new.deny = deny
        new.target_id = target_id
        new.target_type = target_type
        return new



from .activity_change import *
from .activity_update import *
from .guild_profile import *
from .thread_profile import *
from .user import *
from .voice_state import *


__all__ = (
    *activity_change.__all__,
    *activity_update.__all__,
    *guild_profile.__all__,
    *thread_profile.__all__,
    *user.__all__,
    *voice_state.__all__,
)



__all__ = ()

import reprlib, warnings

from scarletio import Compound
from scarletio.web_common import Formdata

from ...bases import maybe_snowflake_pair
from ...core import  GUILDS, STICKERS
from ...emoji import Emoji
from ...http import DiscordHTTPClient, VALID_STICKER_IMAGE_MEDIA_TYPES
from ...sticker import Sticker, StickerPack
from ...utils import MEDIA_TYPE_TO_EXTENSION, get_image_media_type

from ..functionality_helpers import ForceUpdateCache
from ..request_helpers import get_guild_and_id, get_guild_id, get_sticker_and_id, get_sticker_pack_and_id


STICKER_PACK_CACHE = ForceUpdateCache()


def _handle_emoji_representation_parameter_deprecation(instance, tags, emoji_representation):
    """
    Handles the deprecation of the `emoji_representation` parameter.
    
    Parameters
    ----------
    instance : ``Client``
        Client instance used for raising correct warning.
    tags : ``Emoji``, `str`, `iterable` of `str`
        The tags of the sticker.
    emoji_representation : ``Emoji``, `str`, `iterable` of `str`
        The tags of the sticker.
    
    Returns
    -------
    tags : ``Emoji``, `str`, `iterable` of `str`
    """
    if emoji_representation is not ...:
        warnings.warn(
            (
                f'`{instance.__class__.__name__}.sticker_create`\'s `emoji_representation` is deprecated '
                f'and will be removed in 2023 Marc. '
                f'Please use `tags` instead.'
            ),
            FutureWarning,
            stacklevel = 3,
        )
        
        tags = emoji_representation
    
    return tags


def _validate_tags(tags):
    """
    Validates the given tags value.
    
    Parameters
    ----------
    tags : ``Emoji``, `str`, `iterable` of `str`
        The tags of the sticker.
    
    Returns
    -------
    tags : `set` of `str`
        The validated tags.
    
    Raises
    ------
    TypeError
        - If `tags`'s type is incorrect.
    """
    validated_tags = set()
    
    if isinstance(tags, str):
        validated_tags.add(tags)
    
    elif isinstance(tags, Emoji):
        validated_tags.add(tags.name)
    
    elif (getattr(tags, '__iter__', None) is not None):
        for tag in tags:
            if not isinstance(tag, str):
                raise TypeError(
                    f'`tags` can have `str` elements, got {tag.__class__.__name__}; {tag!r}; tags = {tags!r}'
                )
            validated_tags.add(tag)
    
    else:
        raise TypeError(
            f'`tags` can be `str`, `{Emoji.__name__}`, `iterable` of `str`, got {tags.__class__.__name__}; {tags!r}.'
        )
    
    return validated_tags


class ClientCompoundStickerEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    
    async def sticker_get(self, sticker, *, force_update = False):
        """
        Gets an sticker by it's id. If the sticker is already loaded updates it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `int`
            The sticker, who will be requested.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker should be requested even if it supposed to be up to date.
        
        Returns
        -------
        sticker : ``Sticker``
        
        Raises
        ------
        TypeError
            If `sticker` was not given neither as ``Sticker``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        TypeError
            If `sticker` was not given as ``Sticker``, nor as `int`.
        """
        sticker, sticker_id = get_sticker_and_id(sticker)
        if (sticker is not None) and (not sticker.partial) and (not force_update):
            return sticker
        
        data = await self.http.sticker_get(sticker_id)
        if (sticker is None):
            sticker = Sticker.from_data(data)
        else:
            sticker._set_attributes(data)
        
        return sticker
    
    
    async def sticker_pack_get(self, sticker_pack, *, force_update=False):
        """
        Gets the sticker packs. If the sticker-packs are already loaded, updates them.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker_pack : ``StickerPack``, `int`
            The sticker pack' identifier.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker-pack should be requested even if it supposed to be up to date.
        
        Returns
        -------
        sticker_packs : `list` of ``StickerPack``
        
        Raises
        ------
        TypeError
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        sticker_pack, sticker_pack_id = get_sticker_pack_and_id(sticker_pack)
        
        if (sticker_pack is None) or force_update:
            sticker_pack_data = await self.http.sticker_pack_get(sticker_pack_id)
            sticker_pack = StickerPack.from_data(sticker_pack_data, force_update = True)
        
        return sticker_pack
    
    
    async def sticker_pack_get_all(self, *, force_update=False):
        """
        Gets the sticker packs. If the sticker-packs are already loaded, updates them.
        
        This method is a coroutine.
        
        Parameters
        ----------
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker-packs should be requested even if it supposed to be up to date.
        
        Returns
        -------
        sticker_packs : `list` of ``StickerPack``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if force_update or (not STICKER_PACK_CACHE.synced):
            data = await self.http.sticker_pack_get_all()
            sticker_pack_datas = data['sticker_packs'] # Discord pls.
            sticker_packs = [
                StickerPack._create_and_update(sticker_pack_data, force_update = force_update)
                for sticker_pack_data in sticker_pack_datas
            ]
            
            STICKER_PACK_CACHE.set(sticker_packs)
        
        else:
            sticker_packs = STICKER_PACK_CACHE.value
        
        return sticker_packs
    
    
    async def sticker_get_guild(self, sticker, *, force_update = False):
        """
        Gets the specified sticker from the respective guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `tuple`, (`int`, `int`)
            The sticker to get.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker should be requested even if it supposed to be up to date.
        
        Raises
        ------
        TypeError
            If `sticker` is not ``Sticker``, `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if isinstance(sticker, Sticker):
            guild_id = sticker.guild_id
            sticker_id = sticker.id
        else:
            snowflake_pair = maybe_snowflake_pair(sticker)
            if snowflake_pair is None:
                raise TypeError(
                    f'`sticker` can be `{Sticker.__name__}`, `tuple` (`int`, `int`), '
                    f'got {sticker.__class__.__name__}; {sticker!r}.'
                )
            
            guild_id, sticker_id = snowflake_pair
            sticker = STICKERS.get(sticker_id, None)
        
        if (sticker is not None) and (not sticker.partial) and (not force_update):
            return sticker
        
        data = await self.http.sticker_get_guild(guild_id, sticker_id)
        if (sticker is None):
            sticker = Sticker.from_data(data)
        else:
            sticker._set_attributes(data)
        
        return sticker
    
    
    async def sticker_create(
        self, guild, name, image, tags = ..., description = None, *, emoji_representation = ..., reason = None
    ):
        """
        Creates a sticker in the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to create the sticker in.
        name : `str`
            The sticker's name. It's length can be in range [2:32]
        tags : ``Emoji``, `str`, `iterable` of `str`
            The tags of the sticker.
        image : `bytes-like`
            The sticker's image in bytes.
        description : `None`, `str` = `None`, Optional
            The sticker's representation. It's length can be in range [0:100]
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Returns
        -------
        sticker : ``Sticker``
            The created sticker.
        
        Raises
        ------
        TypeError
            - If `str` is neither `str` nor ``Emoji``, `iterable` of `str`.
            - If `guild` is neither ``Guild`` nor `int`.
        ValueError
            If `image`s media type is neither `image/png` nor `application/json`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` is not `str`.
            - If `name`'s length is out of range [2:32].
            - If `description` is neither `None`, `str`.
            - If `description`'s length is out of range [0:100].
        """
        # Handle deprecation.
        tags = _handle_emoji_representation_parameter_deprecation(self, tags, emoji_representation)
        if tags is ...:
            raise TypeError('`tags` parameter is required.')
        
        guild_id = get_guild_id(guild)
        
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(
                    f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                )
        
            name_length = len(name)
            if (name_length < 2) or (name_length > 32):
                raise AssertionError(
                    f'`name` length can be in range [2:32], got {name_length!r}; {name!r}.'
                )
        
        if __debug__:
            if (description is not None):
                if (not isinstance(description, str)):
                    raise AssertionError(
                        f'`description` can be `str`, got {description.__class__.__name__}; {description!r}.'
                    )
                
                description_length = len(description)
                if (description_length > 100):
                    raise AssertionError(
                        f'`description` length can be in range [0:100], got {description_length!r}; {description!r}.'
                    )
        
        if (description is not None) and (not description):
            description = None
        
        tags = _validate_tags(tags)
        
        if __debug__:
            if not isinstance(image, (bytes, bytearray, memoryview)):
                raise TypeError(
                    f'`image` can be `None`, `bytes-like`, got {image.__class__.__name__}; {reprlib.repr(image)}.'
                )
        
        media_type = get_image_media_type(image)
        if media_type not in VALID_STICKER_IMAGE_MEDIA_TYPES:
            raise ValueError(
                f'Invalid `image` type: {media_type}, got {reprlib.repr(image)}.'
            )
        
        extension = MEDIA_TYPE_TO_EXTENSION[media_type]
        
        form_data = Formdata()
        form_data.add_field('name', name)
        
        if (description is not None):
            form_data.add_field('description', description)
        else:
            # If no description is given Discord drops back an unrelated error
            form_data.add_field('description', '')
        
        form_data.add_field('tags', ', '.join(tags))
        
        form_data.add_field('file', image, filename = f'file.{extension}', content_type = media_type)
        
        
        sticker_data = await self.http.sticker_create(guild_id, form_data, reason)
        
        return Sticker.from_data(sticker_data)
    
    
    async def sticker_edit(
        self, sticker, *, name = ..., tags = ..., emoji_representation = ..., description = ..., reason = None
    ):
        """
        Edits the given guild bound sticker,
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `int`
            The respective sticker.
        name : `str`, Optional (keyword only)
            New name of the sticker. It's length can be in range [2:32].
        tags : `str`, ``Emoji``, `iterable` of `str`, Optional (keyword Only)
            The new emoji representation of the sticker. Used as a tag for the sticker.
        description : `None`, `str`, Optional (Keyword only)
            New description for the sticker. It's length can be in range [0:100].
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `sticker` is neither ``Sticker``, nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - Non guild bound sticker cannot be edited.
            - If `name` is not `str`.
            - If `name`'s length is out of range [2:32].
            - If `description` is neither `None`, `str`.
            - If `description`'s length is out of range [0:100].
        """
        # Handle deprecation.
        tags = _handle_emoji_representation_parameter_deprecation(self, tags, emoji_representation)
        
        sticker = await self.sticker_get(sticker)
        
        if __debug__:
            if not sticker.guild_id:
                raise AssertionError(
                    f'Non guild bound sticker cannot be edited, got {sticker!r}.'
                )
        
        if (name is ...):
            name = sticker.name
        else:
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                    )
            
                name_length = len(name)
                if (name_length < 2) or (name_length > 32):
                    raise AssertionError(
                        f'`name` length can be in range [2:32], got {name_length!r}; {name!r}.'
                    )
        
        if (tags is ...):
            tags = sticker.tags
        else:
            tags = _validate_tags(tags)
        
        if (description is ...):
            description = sticker.description
        else:
            if __debug__:
                if (description is not None):
                    if (not isinstance(description, str)):
                        raise AssertionError(
                            f'`description` can be `None`, `str`, got {description.__class__.__name__}; '
                            f'{description!r}.'
                        )
            
                    description_length = len(description)
                    if (description_length > 100):
                        raise AssertionError(
                            f'`description` length can be in range [0:100], got {description_length!r}; '
                            f'{description!r}.'
                        )
            
            if (description is None):
                description = ''
        
        data = {
            'name': name,
            'tags': ', '.join(tags),
            'description': description,
        }
        
        await self.http.sticker_edit(sticker.guild_id, sticker.id, data, reason)
    
    
    async def sticker_delete(self, sticker, *, reason = None):
        """
        Deletes the sticker.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `int`
            The sticker to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `sticker` is neither ``Sticker``, nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            Non guild bound sticker cannot be edited.
        """
        sticker = await self.sticker_get(sticker)
        
        if __debug__:
            if not sticker.guild_id:
                raise AssertionError(
                    f'Non guild bound sticker cannot be edited, got {sticker!r}.'
                )
        
        await self.http.sticker_delete(sticker.guild_id, sticker.id, reason)
    
    
    async def sticker_get_all_guild(self, guild):
        """
        Syncs the given guild's stickers with the wrapper.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's stickers will be synced.
        
        Returns
        -------
        stickers : `list` of ``Sticker``
            The guild's stickers.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        sticker_datas = await self.http.sticker_get_all_guild(guild_id)
        
        if guild is None:
            # Do not create a partial guild, because it would have been garbage collected after leaving the function
            # anyways
            guild = GUILDS.get(guild_id, None)
        
        if guild is None:
            stickers = [Sticker.from_data(sticker_data) for sticker_data in sticker_datas]
        else:
            guild._update_stickers(sticker_datas)
            stickers = [*guild.stickers.values()]
        
        return stickers
    
    
    async def sticker_guild_get_all(self, *position_parameters, **keyword_parameters):
        """
        Deprecated and will be removed in 2023 December. Please use ``.sticker_get_all_guild`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.sticker_guild_get_all` is deprecated and will be removed in 2023 '
                f'December. '
                f'Please use `.sticker_get_all_guild` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return await self.sticker_get_all_guild(*position_parameters, **keyword_parameters)
    
    
    async def sticker_guild_create(self, *position_parameters, **keyword_parameters):
        """
        Deprecated and will be removed in 2023 December. Please use ``.sticker_create`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.sticker_guild_create` is deprecated and will be removed in 2023 December. '
                f'Please use `.sticker_create` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return await self.sticker_create(*position_parameters, **keyword_parameters)
    
    
    async def sticker_guild_edit(self, *position_parameters, **keyword_parameters):
        """
        Deprecated and will be removed in 2023 December. Please use ``.sticker_edit`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.sticker_guild_edit` is deprecated and will be removed in 2023 December. '
                f'Please use `.sticker_edit` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return await self.sticker_edit(*position_parameters, **keyword_parameters)

    
    async def sticker_guild_delete(self, *position_parameters, **keyword_parameters):
        """
        Deprecated and will be removed in 2023 December. Please use ``.sticker_delete`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.sticker_guild_delete` is deprecated and will be removed in 2023 December. '
                f'Please use `.sticker_delete` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return await self.sticker_delete(*position_parameters, **keyword_parameters)
    
    
    async def sticker_guild_get(self, *position_parameters, **keyword_parameters):
        """
        Deprecated and will be removed in 2023 December. Please use ``.sticker_get_guild`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.sticker_guild_get` is deprecated and will be removed in 2023 December. '
                f'Please use `.sticker_get_guild` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return await self.sticker_get_guild(*position_parameters, **keyword_parameters)



__all__ = ()

from ...core import GUILDS


def _add_embedded_activity_state_to_guild_cache(embedded_activity_state):
    """
    Adds the embedded activity to it's guild.
    
    Parameters
    ----------
    embedded_activity_state : ``EmbeddedActivityState``
        The embedded activity to add to it's guild's cache.
    """
    try:
        guild = GUILDS[embedded_activity_state.guild_id]
    except KeyError:
        pass
    else:
        embedded_activity_states = guild.embedded_activity_states
        if (embedded_activity_states is None):
            embedded_activity_states = set()
            guild.embedded_activity_states = embedded_activity_states
        
        embedded_activity_states.add(embedded_activity_state)


def _remove_embedded_activity_state_from_guild_cache(embedded_activity_state):
    """
    Adds the embedded activity to it's guild.
    
    Parameters
    ----------
    embedded_activity_state : ``EmbeddedActivityState``
        The embedded activity to add to it's guild's cache.
    """
    try:
        guild = GUILDS[embedded_activity_state.guild_id]
    except KeyError:
        pass
    else:
        embedded_activity_states = guild.embedded_activity_states
        if (embedded_activity_states is not None):
            try:
                embedded_activity_states.remove(embedded_activity_state)
            except KeyError:
                pass
            else:
                if not embedded_activity_state:
                    guild.embedded_activity_states = None



__all__ = ()

from ...field_parsers import int_parser_factory, nullable_entity_array_parser_factory, nullable_string_parser_factory
from ...field_putters import int_putter_factory, nullable_entity_array_putter_factory, url_optional_putter_factory
from ...field_validators import (
    int_conditional_validator_factory, nullable_entity_array_validator_factory, url_optional_validator_factory
)

from ..guild.fields import parse_id, parse_name, put_id_into, put_name_into, validate_id, validate_name
from ..guild_widget_channel import GuildWidgetChannel
from ..guild_widget_user import GuildWidgetUser

# approximate_online_count

parse_approximate_online_count = int_parser_factory('presence_count', 0)
put_approximate_online_count_into = int_putter_factory('presence_count')
validate_approximate_online_count = int_conditional_validator_factory(
    'approximate_online_count',
    0,
    lambda approximate_online_count : approximate_online_count >= 0,
    '>= 0',
)

# channels

parse_channels = nullable_entity_array_parser_factory('channels', GuildWidgetChannel)
put_channels_into = nullable_entity_array_putter_factory('channels', GuildWidgetChannel)
validate_channels = nullable_entity_array_validator_factory('channels', GuildWidgetChannel)

# invite_url

parse_invite_url = nullable_string_parser_factory('instant_invite')
put_invite_url_into = url_optional_putter_factory('instant_invite')
validate_invite_url = url_optional_validator_factory('invite_url')

# users

parse_users = nullable_entity_array_parser_factory('members', GuildWidgetUser)
put_users_into = nullable_entity_array_putter_factory('members', GuildWidgetUser)
validate_users = nullable_entity_array_validator_factory('members', GuildWidgetUser)



__all__ = ()

from ...auto_moderation import AutoModerationAction, AutoModerationEventType, AutoModerationRuleTriggerType
from ...auto_moderation.trigger_metadata.utils import try_get_auto_moderation_trigger_metadata_type_from_data

from ..audit_log_change import AuditLogChange

from .shared import _convert_preinstanced, convert_nothing, convert_snowflake, convert_snowflake_array


def convert_auto_moderation_rule_actions(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        if before:
            before = (*(AutoModerationAction.from_data(data) for data in before),)
        else:
            before = None
    
    after = data.get('new_value', None)
    if (after is not None):
        if after:
            after = (*(AutoModerationAction.from_data(data) for data in after),)
        else:
            after = None
    
    return AuditLogChange('actions', before, after)


def convert_auto_moderation_rule_event_type(name, data):
    return _convert_preinstanced('event_type', data, AutoModerationEventType)


def convert_snowflake_array__excluded_channel_ids(name, data):
    return convert_snowflake_array('excluded_channel_ids', data)


def convert_snowflake_array__excluded_role_ids(name, data):
    return convert_snowflake_array('excluded_role_ids', data)


def convert_auto_moderation_trigger_entity_metadata(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        metadata_type = try_get_auto_moderation_trigger_metadata_type_from_data(before)
        if metadata_type is None:
            before = None
        else:
            before = metadata_type.from_data(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        metadata_type = try_get_auto_moderation_trigger_metadata_type_from_data(after)
        if metadata_type is None:
            after = None
        else:
            after = metadata_type.from_data(after)
    
    return AuditLogChange('trigger_metadata', before, after)


def convert_auto_moderation_rule_trigger_type(name, data):
    return _convert_preinstanced('trigger_type', data, AutoModerationRuleTriggerType)


def convert_keywords(name, data):
    keywords = (*data['new_value'],)
    
    if name.startswith('$add'):
        before = None
        after = keywords
    else:
        before = keywords
        after = None
    
    return AuditLogChange('keywords', before, after)


def convert_regex_patterns(name, data):
    regex_patterns = (*data['new_value'],)
    
    if name.startswith('$add'):
        before = None
        after = regex_patterns
    else:
        before = regex_patterns
        after = None
    
    return AuditLogChange('regex_patterns', before, after)


def convert_excluded_keywords(name, data):
    excluded_keywords = (*data['new_value'],)
    
    if name.startswith('$add'):
        before = None
        after = excluded_keywords
    else:
        before = excluded_keywords
        after = None
    
    return AuditLogChange('excluded_keywords', before, after)


AUTO_MODERATION_RULE_CONVERTERS = {
    '$add_keyword_filter': convert_keywords,
    '$remove_keyword_filter': convert_keywords,
    '$add_regex_patterns': convert_regex_patterns,
    '$remove_regex_patterns': convert_regex_patterns,
    '$add_allow_list': convert_excluded_keywords,
    '$remove_allow_list': convert_excluded_keywords,
    'actions': convert_auto_moderation_rule_actions,
    'creator_id': convert_snowflake,
    'enabled': convert_nothing,
    'event_type': convert_auto_moderation_rule_event_type,
    'exempt_channels': convert_snowflake_array__excluded_channel_ids,
    'exempt_roles': convert_snowflake_array__excluded_role_ids,
    'name': convert_nothing,
    'trigger_metadata': convert_auto_moderation_trigger_entity_metadata,
    'trigger_type': convert_auto_moderation_rule_trigger_type,
}



__all__ = ('WelcomeScreenChannel',)

import warnings

from scarletio import BaseMethodDescriptor, RichAttributeErrorBaseType

from ...channel import Channel, ChannelType, create_partial_channel_from_id

from .fields import (
    parse_channel_id, parse_description, parse_emoji, put_channel_id_into, put_description_into, put_emoji_into,
    validate_channel_id, validate_description, validate_emoji
)


class WelcomeScreenChannel(RichAttributeErrorBaseType):
    """
    Represents a featured channel by a welcome screen.
    
    Attributes
    ----------
    channel_id : `int`
        The represented channel's identifier.
    description : `None`, `str`
        The channel's short description.
    emoji : `None`, ``Emoji``
        An emoji displayed before the `description`.
    """
    __slots__ = ('channel_id', 'description', 'emoji', )
    
    def __new__(cls, *, channel_id = ..., description = ..., emoji = ...):
        """
        Creates a new welcome channel.
        
        Parameters
        -----------
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The represented channel or its identifier.
        description : `None`, `str`, Optional (Keyword only)
            The channel's short description.
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            An emoji displayed before the `description`.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_channel_id(channel_id)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # Construct
        self = object.__new__(cls)
        self.channel_id = channel_id
        self.description = description
        self.emoji = emoji
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new welcome channel instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Welcome channel data.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.channel_id = parse_channel_id(data)
        self.description = parse_description(data)
        self.emoji = parse_emoji(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the welcome channel to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `str`
        """
        data = {}
        put_channel_id_into(self.channel_id, data, defaults)
        put_description_into(self.description, data, defaults)
        put_emoji_into(self.emoji, data, defaults)
        return data
    
    
    @BaseMethodDescriptor
    def custom(cls, base, *, channel = ..., channel_id = ..., **keyword_parameters):
        """
        Deprecated and will be removed in 2023 April. Please use ``.__new__`` or ``.copy_with`` respectively.
        """
        warnings.warn(
            (
                f'`{cls.__name__}.custom` is deprecated and will be removed in 2023 April. '
                f'Please use `.__new__` or `.copy_with` respectively.'
            ),
            FutureWarning,
            stacklevel = 3,
        )
        
        if channel is not ...:
            warnings.warn(
                (
                    f'`{cls.__name__}.custom`\'s `channel` parameters is deprecated. Please use `channel_id` instead.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
            channel_id = channel
        
        if base is None:
            return cls.__new__(cls, channel_id = channel_id, **keyword_parameters)
        
        return base.copy_with(channel_id = channel_id, **keyword_parameters)
    
        
    
    def __repr__(self):
        """Returns the welcome channel's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # channel_id
        repr_parts.append(' channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        # description
        description = self.description
        if (description is not None):
            repr_parts.append(', description = ')
            repr_parts.append(repr(description))
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            repr_parts.append(', emoji = ')
            repr_parts.append(repr(emoji))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the welcome channel's hash."""
        hash_value = 0
        
        # channel_id
        hash_value ^= self.channel_id
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= hash(emoji)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two welcome channels are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the welcome channel.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.channel_id = self.channel_id
        new.description = self.description
        new.emoji = self.emoji
        return new
    
    
    def copy_with(self, *, channel_id = ..., description = ..., emoji = ...):
        """
        Creates a new welcome channel with the given fields.
        
        Parameters
        -----------
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The represented channel or its identifier.
        description : `None`, `str`, Optional (Keyword only)
            The channel's short description.
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            An emoji displayed before the `description`.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # Construct
        new = object.__new__(type(self))
        new.channel_id = channel_id
        new.description = description
        new.emoji = emoji
        return new
    
    
    @property
    def channel(self):
        """
        Returns the welcome channel's respective channel.
        
        Returns
        -------
        channel : ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.guild_text, 0)



from .allow import *
from .deny import *
from .target import *
from .target_id import *
from .target_type import *

__all__ = (
    *allow.__all__,
    *deny.__all__,
    *target.__all__,
    *target_id.__all__,
    *target_type.__all__,
)



__all__ = ('ApplicationCommand',)

import warnings

from ...bases import DiscordEntity
from ...core import APPLICATION_COMMANDS, GUILDS
from ...localization.helpers import get_localized_length
from ...localization.utils import hash_locale_dictionary
from ...permission import Permission
from ...utils import DATETIME_FORMAT_CODE, id_to_datetime

from ..application_command_option import ApplicationCommandOption
from ..helpers import with_translation

from .constants import APPLICATION_COMMAND_OPTIONS_MAX
from .fields import (
    parse_allow_in_dm, parse_application_id, parse_description, parse_description_localizations, parse_guild_id,
    parse_id, parse_name, parse_name_localizations, parse_nsfw, parse_options, parse_required_permissions,
    parse_target_type, parse_version, put_allow_in_dm_into, put_application_id_into, put_description_into,
    put_description_localizations_into, put_guild_id_into, put_id_into, put_name_into, put_name_localizations_into,
    put_nsfw_into, put_options_into, put_required_permissions_into, put_target_type_into, put_version_into,
    validate_allow_in_dm, validate_description, validate_description_localizations, validate_name,
    validate_name_localizations, validate_nsfw, validate_options, validate_required_permissions, validate_target_type
)
from .preinstanced import APPLICATION_COMMAND_CONTEXT_TARGET_TYPES, ApplicationCommandTargetType


class ApplicationCommand(DiscordEntity, immortal = True):
    """
    Represents a Discord slash command.
    
    Attributes
    ----------
    allow_in_dm : `bool`
        Whether the command can be used in private (dm) channels.
    
    application_id : `int`
        The application command's application's id.
    
    description : `None`, `str`
        The command's description. It's length can be in range [2:100].
        
        Set as `None` for context commands.
    
    description_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localized descriptions of the application command.
        
        Set as `None` for context commands.
    
    guild_id : `int`
        The guild's identifier to which the command is bound to.
        
        Set as `0` if the command is global.
    
    id : `int`
        The application command's id.
    
    name : `str`
        The name of the command. It's length can be in range [1:32].
    
    name_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localized names of the application command.
    
    nsfw : `bool`
        Whether the application command is only allowed in nsfw channels.
    
    options : `None`, `tuple` of ``ApplicationCommandOption``
        The parameters of the command. It's length can be in range [0:25]. If would be set as empty list, instead is
        set as `None`.
    
    required_permissions : ``Permission``
        The required permissions to use the application command inside of a guild.
    
    target_type : ``ApplicationCommandTargetType``
        The application command target's type describing where it shows up.
    
    version : `int`
        The time when the command was last edited in snowflake.
    
    Notes
    -----
    Application command instances are weakreferable.
    """
    __slots__ = (
        'allow_in_dm', 'application_id', 'description', 'description_localizations', 'guild_id', 'name',
        'name_localizations', 'nsfw', 'options', 'required_permissions', 'target_type', 'version'
    )
    
    def __new__(
        cls,
        name,
        description = None,
        *,
        allow_in_dm = ...,
        description_localizations = ...,
        name_localizations = ...,
        nsfw = ...,
        options = ...,
        required_permissions = ...,
        target_type = ...,
    ):
        """
        Creates a new application command with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the command. It's length can be in range [1:32].
        
        description : `None`, `str` = `None`, Optional
            The command's description. It's length can be in range [2:100].
            
            Defaults to the `name` parameter if not given.
        
        allow_in_dm : `None`, `bool`, Optional (Keyword only)
            Whether the command can be used in private channels (dm).
            
            Defaults to `True`
        
        description_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`), Optional (Keyword only)
            Localized descriptions of the application command.
        
        name_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`), Optional (Keyword only)
            Localized names of the application command.
        
        nsfw : `None`, `bool`, Optional (Keyword only)
            Whether the application command is allowed in nsfw channels.
        
        options : `None`, `iterable` of ``ApplicationCommandOption``, Optional (Keyword only)
            The parameters of the command. It's length can be in range [0:25].
        
        required_permissions : `None`, ``Permission``, `int`, Optional (Keyword only)
            The required permissions to use the application command inside of a guild.
        
        target_type : `None`, `int`, ``ApplicationCommandTargetType``, Optional (Keyword only)
            The application command's target type.
            
            Defaults to `ApplicationCommandTargetType.chat`.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # allow_in_dm
        if allow_in_dm is ...:
            allow_in_dm = True
        else:
            allow_in_dm = validate_allow_in_dm(allow_in_dm)
        
        # description
        description = validate_description(description)
        
        # description_localizations
        if description_localizations is ...:
            description_localizations = None
        else:
            description_localizations = validate_description_localizations(description_localizations)
        
        # name
        name = validate_name(name)
        
        # name_localizations
        if name_localizations is ...:
            name_localizations = None
        else:
            name_localizations = validate_name_localizations(name_localizations)
        
        # nsfw
        if nsfw is ...:
            nsfw = False
        else:
            nsfw = validate_nsfw(nsfw)
        
        # options
        if options is ...:
            options = None
        else:
            options = validate_options(options)
        
        # required_permissions
        if required_permissions is ...:
            required_permissions = Permission()
        else:
            required_permissions = validate_required_permissions(required_permissions)
        
        # target_type
        if target_type is ...:
            target_type = ApplicationCommandTargetType.chat
        else:
            target_type = validate_target_type(target_type)
        
        
        # Post checks
        if (target_type in APPLICATION_COMMAND_CONTEXT_TARGET_TYPES):
            # Context commands cannot have description and options, so we clear them.
            description = None
            description_localizations = None
            options = None
        
        else:
            # For non context commands description is required.
            if (description is None):
                description = validate_description(name)
        
        # Construct
        self = object.__new__(cls)
        self.application_id = 0
        self.name = name
        self.name_localizations = name_localizations
        self.nsfw = nsfw
        self.description = description
        self.description_localizations = description_localizations
        self.guild_id = 0
        self.id = 0
        self.allow_in_dm = allow_in_dm
        self.options = options
        self.required_permissions = required_permissions
        self.target_type = target_type
        self.version = 0
        return self
    
    
    @classmethod
    def _create_empty(cls, application_command_id, application_id):
        """
        Creates an empty application command with the default attributes set.
        
        Parameters
        ----------
        application_command_id : `int`
            The application command's identifier.
        application_id : `int`
            The application command's owner application's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = application_command_id
        self.allow_in_dm = True
        self.application_id = application_id
        self.description = None
        self.description_localizations = None
        self.guild_id = 0
        self.name = ''
        self.name_localizations = None
        self.nsfw = False
        self.options = None
        self.required_permissions = Permission()
        self.target_type = ApplicationCommandTargetType.none
        self.version = 0
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new Application command from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received application command data.
        
        Returns
        -------
        self : `instance<cls>`
            The created application command instance.
        """
        application_command_id = parse_id(data)
        try:
            self = APPLICATION_COMMANDS[application_command_id]
        except KeyError:
            self = cls._create_empty(application_command_id, parse_application_id(data))
            
            # guild_id
            self.guild_id = parse_guild_id(data)
            
            APPLICATION_COMMANDS[application_command_id] = self
        
        self._update_attributes(data)
        return self
    
    
    @classmethod
    def _from_edit_data(cls, data, application_command_id, application_id):
        """
        Creates an application command with the given parameters after an application command edition took place.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application command data returned by it's ``.to_data`` method.
        application_command_id : `int`
            The unique identifier number of the newly created application command.
        application_id : `int`
            The new application identifier number of the newly created application command.
        
        Returns
        -------
        self : `instance<cls>`
            The newly created or updated application command.
        """
        try:
            self = APPLICATION_COMMANDS[application_command_id]
        except KeyError:
            self = cls._create_empty(application_command_id, application_id)
            APPLICATION_COMMANDS[application_command_id] = self
        
        self._update_attributes(data)
        
        return self
    
    
    def to_data(self, * , defaults = False, include_internals = True):
        """
        Converts the application command to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        put_allow_in_dm_into(self.allow_in_dm, data, defaults)
        put_description_into(self.description, data, defaults)
        put_description_localizations_into(self.description_localizations, data, defaults)
        put_name_into(self.name, data, defaults)
        put_name_localizations_into(self.name_localizations, data, defaults)
        put_nsfw_into(self.nsfw, data, defaults)
        put_options_into(self.options, data, defaults)
        put_required_permissions_into(self.required_permissions, data, defaults)
        put_target_type_into(self.target_type, data, defaults)
        
        if include_internals:
            put_application_id_into(self.application_id, data, defaults)
            put_guild_id_into(self.guild_id, data, defaults)
            put_id_into(self.id, data, defaults)
            put_version_into(self.version, data, defaults)
        
        return data
    
    
    def _update_attributes(self, data):
        """
        Updates the application command with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received application command data.
        """
        self.allow_in_dm = parse_allow_in_dm(data)
        self.description = parse_description(data)
        self.description_localizations = parse_description_localizations(data)
        self.name = parse_name(data)
        self.name_localizations = parse_name_localizations(data)
        self.nsfw = parse_nsfw(data)
        self.options = parse_options(data)
        self.required_permissions = parse_required_permissions(data)
        self.target_type = parse_target_type(data)
        self.version = parse_version(data)

    
    def _difference_update_attributes(self, data):
        """
        Updates the application command with the given data and returns the updated attributes in a dictionary with the
        attribute names as the keys and their old value as the values.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received application command data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `object`) items
            The updated attributes.
            
            Every item in the returned dict is optional and can contain the following ones:
            
            +---------------------------+---------------------------------------------------+
            | Keys                      | Values                                            |
            +===========================+===================================================+
            | description               | `None`, `str`                                     |
            +---------------------------+---------------------------------------------------+
            | description_localizations | `None`, `dict` of (``Locale``, `str`) items       |
            +---------------------------+---------------------------------------------------+
            | allow_in_dm               | `bool`                                            |
            +---------------------------+---------------------------------------------------+
            | name                      | `str`                                             |
            +---------------------------+---------------------------------------------------+
            | name_localizations        | `None`, `dict` of (``Locale``, `str`) items       |
            +---------------------------+---------------------------------------------------+
            | nsfw                      | `bool`                                            |
            +---------------------------+---------------------------------------------------+
            | options                   | `None`, `list` of ``ApplicationCommandOption``    |
            +---------------------------+---------------------------------------------------+
            | required_permissions      | ``Permission``                                    |
            +---------------------------+---------------------------------------------------+
            | target_type               | ``ApplicationCommandTargetType``                  |
            +---------------------------+---------------------------------------------------+
            | version                   | `int`                                             |
            +---------------------------+---------------------------------------------------+
        """
        old_attributes = {}
        
        # allow_in_dm
        allow_in_dm = parse_allow_in_dm(data)
        if self.allow_in_dm != allow_in_dm:
            old_attributes['allow_in_dm'] = self.allow_in_dm
            self.allow_in_dm = allow_in_dm
        
        # description
        description = parse_description(data)
        if self.description != description:
            old_attributes['description'] = self.description
            self.description = description
        
        # description_localizations
        description_localizations = parse_description_localizations(data)
        if self.description_localizations != description_localizations:
            old_attributes['description_localizations'] = self.description_localizations
            self.description_localizations = description_localizations
        
        # name
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # name_localizations
        name_localizations = parse_name_localizations(data)
        if self.name_localizations != name_localizations:
            old_attributes['name_localizations'] = self.name_localizations
            self.name_localizations = name_localizations
        
        # nsfw
        nsfw = parse_nsfw(data)
        if self.nsfw != nsfw:
            old_attributes['nsfw'] = self.nsfw
            self.nsfw = nsfw
        
        # options
        options = parse_options(data)
        if self.options != options:
            old_attributes['options'] = self.options
            self.options = options
        
        # required_permissions
        required_permissions = parse_required_permissions(data)
        if self.required_permissions != required_permissions:
            old_attributes['required_permissions'] = self.required_permissions
            self.required_permissions = required_permissions
        
        # target_type
        target_type = parse_target_type(data)
        if (self.target_type is not target_type):
            old_attributes['target_type'] = self.target_type
            self.target_type = target_type
        
        # version
        version = parse_version(data)
        if self.version != version:
            old_attributes['version'] = self.version
            self.version = version
        
        return old_attributes
    
    
    def __repr__(self):
        """Returns the application command's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # if the application command is partial, mention that, else add  `.id` and `.application_id` fields.
        if self.partial:
            repr_parts.append(' (partial)')
        
        else:
            # id
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            
            # application_id
            repr_parts.append(', application_id = ')
            repr_parts.append(repr(self.application_id))
            
            # guild_id
            guild_id = self.guild_id
            if guild_id:
                repr_parts.append(', guild_id = ')
                repr_parts.append(repr(guild_id))
        
        # Required fields are `.name` and `.type`
        
        # name
        repr_parts.append(', name = ')
        repr_parts.append(repr(self.name))
        
        # target_type
        target_type = self.target_type
        if (target_type is not ApplicationCommandTargetType.none):
            repr_parts.append(', target_type = ')
            repr_parts.append(target_type.name)
            repr_parts.append(' (')
            repr_parts.append(repr(target_type.value))
            repr_parts.append(')')
        
        # Extra fields: `.description`, `.options`, `.allow_in_dm`, `.required_permissions`, `.nsfw`,
        # `.name_localizations`, `.description_localizations`
        
        # description
        description = self.description
        if (description is not None):
            repr_parts.append(', description = ')
            repr_parts.append(repr(self.description))
        
        # allow_in_dm
        if not self.allow_in_dm:
            repr_parts.append(', allow_in_dm = False')
        
        required_permissions = self.required_permissions
        if required_permissions:
            repr_parts.append(', required_permissions = ')
            repr_parts.append(required_permissions.__format__('d'))
        
        if self.nsfw:
            repr_parts.append(', nsfw = True')
        
        # options
        options = self.options
        if (options is not None):
            repr_parts.append(', options = [')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            repr_parts.append(', name_localizations = ')
            repr_parts.append(repr(name_localizations))
        
        # description_localizations
        description_localizations = self.description_localizations
        if (description_localizations is not None):
            repr_parts.append(', description_localizations = ')
            repr_parts.append(repr(description_localizations))
        
        # Ignore extra fields: `.version`
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the application's hash value."""
        application_command_id = self.id
        if application_command_id:
            return application_command_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Hashes the fields of the application command.
        
        Called by ``.__hash__` when the application command is partial.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # id
        # non-partial field
        
        # allow_in_dm
        hash_value ^= self.allow_in_dm << 1
        
        # application_id
        # non-partial field
        
        # description
        # Do not hash `.description` if same as `.name`
        description = self.description
        if (description is not None) and (description != self.name):
            hash_value ^= hash(description)
        
        # description_localizations
        # Do not hash `.description_localizations` if same as `.name_localizations`
        description_localizations = self.description_localizations
        if (description_localizations is not None) and (description_localizations != self.name_localizations):
            hash_value ^= hash_locale_dictionary(description_localizations)
        
        # guild_id
        # non-partial field
        
        # name
        hash_value ^= hash(self.name)
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            hash_value ^= hash_locale_dictionary(name_localizations)
        
        # nsfw
        hash_value ^= self.nsfw << 2
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options) << 3
            
            for option in options:
                hash_value ^= hash(option)
        
        # required_permissions
        required_permissions = self.required_permissions
        if required_permissions:
            hash_value ^= hash(required_permissions << 7)
        
        # target_type
        hash_value ^= self.target_type.value << 11
        
        # version
        hash_value ^= self.version << 15
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two application commands are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two application commands are different."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two instances are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `type<self>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # If both entity is not partial, leave instantly by comparing id.
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self_id == other_id:
                return True
            
            return False
        
        # allow_in_dm
        if self.allow_in_dm != other.allow_in_dm:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # description_localizations
        if self.description_localizations != other.description_localizations:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # name_localizations
        if self.name_localizations != other.name_localizations:
            return False
        
        # nsfw
        if self.nsfw != other.nsfw:
            return False
        
        # required_permissions
        if self.required_permissions != other.required_permissions:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # target_type
        if (self.target_type is not other.target_type):
            return False
        
        return True
    
    
    def __format__(self, code):
        """
        Formats the application command in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        application_command : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import ApplicationCommand
        >>> application_command = ApplicationCommand('cake-lover', 'Sends a random cake recipe OwO')
        >>> application_command
        <ApplicationCommand partial name = 'cake-lover', description = 'Sends a random cake recipe OwO'>
        >>> # no code stands for `application_command.name`.
        >>> f'{application_command}'
        'CakeLover'
        >>> # 'd' stands for display name.
        >>> f'{application_command:d}'
        'cake-lover'
        >>> # 'm' stands for mention.
        >>> f'{application_command:m}'
        '</cake-lover:0>'
        >>> # To mention a sub command, use @sub-command's name.
        >>> f'{application_command:m@eat}'
        '</cake-lover eat:0>'
        >>> # 'c' stands for created at.
        >>> f'{application_command:c}'
        '2021-01-03 20:17:36'
        >>> # 'e' stands for edited at.
        >>> f'{application_command:e}'
        'never'
        ```
        """
        if not code:
            return self.name
        
        if code == 'm' or code == 'm@':
            return self.mention
        
        if code.startswith('m@'):
            return self.mention_with(code[2:])
        
        if code == 'd':
            return self.display_name
        
        if code == 'c':
            return format(self.created_at, DATETIME_FORMAT_CODE)
        
        if code == 'e':
            edited_at = self.edited_at
            if edited_at is None:
                edited_at = 'never'
            else:
                edited_at = format(edited_at, DATETIME_FORMAT_CODE)
            return edited_at
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"c"!r}, {"d"!r}, {"e"!r}, {"m"!r}, {"m@..."!r}.'
        )
    
    
    def __len__(self):
        """Returns the application command's length."""
        length = 0
        
        # description & description_localizations
        length += get_localized_length(self.description, self.description_localizations)
        
        # name & name_localizations
        length += get_localized_length(self.name, self.name_localizations)
        
        # options
        options = self.options
        if (options is not None):
            for option in options:
                length += len(option)
        
        return length
    
    
    def apply_translation(self, translation_table, replace = False):
        """
        Applies translation from the given nested dictionary to the application command.
        
        Parameters
        ----------
        translation_table : `None`, `dict` of ((``Locale``, `str`),
                (`None`, `dict` (`str`, (`None`, `str`)) items)) items
            Translation table to pull localizations from.
        replace : `bool` = `False`, Optional
            Whether actual translation should be replaced.
        
        Raises
        ------
        RuntimeError
            - If the application command is not partial.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.apply_translation` is deprecated and will be removed in 2023 Jul. '
                f'Please use `.with_translation` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        if not self.partial:
            raise RuntimeError(
                f'{self.__class__.__name__}.add_option` can be only called on partial '
                f'`{self.__class__.__name__}`-s, but was called on {self!r}.'
            )
        
        if translation_table is None:
            return
        
        # description
        self.description_localizations = with_translation(
            self.description,
            self.description_localizations,
            translation_table,
            replace,
        )
        
        # name
        self.name_localizations = with_translation(
            self.name,
            self.name_localizations,
            translation_table,
            replace,
        )
        
        # options
        options = self.options
        if (options is not None):
            for option in options:
                option.apply_translation(translation_table, replace)
    
    
    def add_option(self, option):
        """
        Adds a new option to the application command.
        
        Parameters
        ----------
        option : ``ApplicationCommandOption``
            The option to add.
        
        Returns
        -------
        self : `instance<type<self>>`
        
        Raises
        ------
        RuntimeError
            - If the entity is not partial.
        TypeError
            - If `option` is not ``ApplicationCommandOption``.
        """
        warnings.warn(
            f'`{self.__class__.__name__}.add_option` is deprecated and will be removed in 2023 Jul.',
            FutureWarning,
            stacklevel = 2,
        )
        
        if not self.partial:
            raise RuntimeError(
                f'{self.__class__.__name__}.add_option` can be only called on partial '
                f'`{self.__class__.__name__}`-s, but was called on {self!r}.'
            )
        
        if not isinstance(option, ApplicationCommandOption):
            raise TypeError(
                f'`option` can be `{ApplicationCommandOption.__name__}`, got '
                f'{option.__class__.__name__}; {option!r}.'
            )
        
        self.options = (*self.iter_options(), option)[:APPLICATION_COMMAND_OPTIONS_MAX]
        return self
    
    
    def copy(self):
        """
        Copies the application command.
        
        The copy is always a partial application command.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        # allow_in_dm
        new.allow_in_dm = self.allow_in_dm
        
        # application_id
        new.application_id = 0
        
        # description
        new.description = self.description
        
        # description_localizations
        description_localizations = self.description_localizations
        if (description_localizations is not None):
            description_localizations = description_localizations.copy()
        new.description_localizations = description_localizations
        
        # guild_id
        new.guild_id = 0
        
        # id
        new.id = 0
        
        # name
        new.name = self.name
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            name_localizations = name_localizations.copy()
        new.name_localizations = name_localizations
        
        # nsfw
        new.nsfw = self.nsfw
        
        # options
        options = self.options
        if (options is not None):
            options = (*(option.copy() for option in options),)
        new.options = options
        
        # required_permissions
        new.required_permissions = self.required_permissions
        
        # target_type
        new.target_type = self.target_type
        
        # version
        new.version = 0
        
        return new
    
    
    def copy_with(
        self,
        *,
        allow_in_dm = ...,
        description = ...,
        description_localizations = ...,
        name = ...,
        name_localizations = ...,
        nsfw = ...,
        options = ...,
        required_permissions = ...,
        target_type = ...,
    ):        
        """
        Copies the application command with the given fields.
        
        Parameters
        ----------
        allow_in_dm : `None`, `bool`, Optional (Keyword only)
            Whether the command can be used in private channels (dm).
        
        description : `None`, `str` = `None`, Optional
            The command's description. It's length can be in range [2:100].
        
        description_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`), Optional (Keyword only)
            Localized descriptions of the application command.
        
        name : `str`, Optional (Keyword only)
            The name of the command. It's length can be in range [1:32].
        
        name_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`), Optional (Keyword only)
            Localized names of the application command.
        
        nsfw : `None`, `bool`, Optional (Keyword only)
            Whether the application command is allowed in nsfw channels.
        
        options : `None`, `iterable` of ``ApplicationCommandOption``, Optional (Keyword only)
            The parameters of the command. It's length can be in range [0:25].
        
        required_permissions : `None`, ``Permission``, `int`, Optional (Keyword only)
            The required permissions to use the application command inside of a guild.
        
        target_type : `None`, `int`, ``ApplicationCommandTargetType``, Optional (Keyword only)
            The application command's target type.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # allow_in_dm
        if allow_in_dm is ...:
            allow_in_dm = self.allow_in_dm
        else:
            allow_in_dm = validate_allow_in_dm(allow_in_dm)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # description_localizations
        if description_localizations is ...:
            description_localizations = self.description_localizations
            if (description_localizations is not None):
                description_localizations = description_localizations.copy()
        else:
            description_localizations = validate_description_localizations(description_localizations)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # name_localizations
        if name_localizations is ...:
            name_localizations = self.name_localizations
            if (name_localizations is not None):
                name_localizations = name_localizations.copy()
        else:
            name_localizations = validate_name_localizations(name_localizations)
        
        # nsfw
        if nsfw is ...:
            nsfw = self.nsfw
        else:
            nsfw = validate_nsfw(nsfw)
        
        # options
        if options is ...:
            options = self.options
            if (options is not None):
                options = (*(option.copy() for option in options),)
        else:
            options = validate_options(options)
        
        # required_permissions
        if required_permissions is ...:
            required_permissions = self.required_permissions
        else:
            required_permissions = validate_required_permissions(required_permissions)
        
        # target_type
        if target_type is ...:
            target_type = ApplicationCommandTargetType.chat
        else:
            target_type = validate_target_type(target_type)
        
        
        # Post checks
        if (target_type in APPLICATION_COMMAND_CONTEXT_TARGET_TYPES):
            # Context commands cannot have description and options, so we clear them.
            description = None
            description_localizations = None
            options = None
        
        else:
            # For non context commands description is required.
            if (description is None):
                description = validate_description(name)
    
        # Construct
        new = object.__new__(type(self))
        new.application_id = 0
        new.name = name
        new.name_localizations = name_localizations
        new.nsfw = nsfw
        new.description = description
        new.description_localizations = description_localizations
        new.guild_id = 0
        new.id = 0
        new.allow_in_dm = allow_in_dm
        new.options = options
        new.required_permissions = required_permissions
        new.target_type = target_type
        new.version = 0
        return new
    
    
    def with_translation(self, translation_table, replace = False):
        """
        Returns a new application command with the given translation table applied.
        
        Parameters
        ----------
        translation_table : `None`, `dict` of ((``Locale``, `str`),
                (`None`, `dict` (`str`, (`None`, `str`)) items)) items
            Translation table to pull localization. from.
        replace : `bool` = `False`, Optional
            Whether actual translation should be replaced.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        if translation_table is None:
            return self
        
        new = self.copy()
        
        # description_localizations
        new.description_localizations = with_translation(
            new.description,
            new.description_localizations,
            translation_table,
            replace,
        )
        
        # name_localizations
        new.name_localizations = with_translation(
            new.name,
            new.name_localizations,
            translation_table,
            replace,
        )
        
        # options
        options = new.options
        if (options is not None):
            new.options = (*(option.with_translation(translation_table, replace) for option in options),)
        
        return new
    
    
    def mention_sub_command(self, *sub_command_names):
        """
        Returns the application command's mention extended with the given sub-command names..
        
        Parameters
        ----------
        *sub_command_names : `str`
            The sub commands' names to mention.
        
        Returns
        -------
        mention : `str`
        """
        mention_parts = ['</', self.name]
        
        for sub_command_name in sub_command_names:
            mention_parts.append(' ')
            mention_parts.append(sub_command_name)
        
        mention_parts.append(':')
        mention_parts.append(str(self.id))
        mention_parts.append('>')
        
        return ''.join(mention_parts)
    
    
    def mention_with(self, with_):
        """
        Returns the application command's mention with the added string.
        
        Parameters
        ----------
        with_ : `str`
            Additional string to mention the command with. It should be sub commands' name.
        
        Returns
        -------
        mention : `str`
        """
        return f'</{self.name} {with_}:{self.id}>'
    
    
    @property
    def mention(self):
        """
        Returns the application command's mention.
        
        Returns
        -------
        mention : `str`
        """
        return f'</{self.name}:{self.id}>'
    
    
    @property
    def display_name(self):
        """
        Returns the application command's display name.
        
        Returns
        -------
        display_name : `str`
        """
        return self.name.lower().replace('_', '-')
    
    
    @property
    def edited_at(self):
        """
        Returns when the command was last edited / modified. If the command was not edited yet, returns `None`.
        
        Returns
        -------
        edited_at : `None`, `edited_at`
        """
        version = self.version
        if version:
            return id_to_datetime(version)
    
    
    @property
    def partial(self):
        """
        Returns whether the application command is partial.
        
        Returns
        -------
        partial : `bool`
        """
        return not self.id
    
    
    def is_context_command(self):
        """
        Returns whether the application command is a context command.
        
        Returns
        -------
        is_context_command : `bool`
        """
        return (self.target_type in APPLICATION_COMMAND_CONTEXT_TARGET_TYPES)
    
    
    def is_slash_command(self):
        """
        Returns whether the application command is a slash command.
        
        Returns
        -------
        is_slash_command : `bool`
        """
        return (self.target_type is ApplicationCommandTargetType.chat)
    
    
    @property
    def guild(self):
        """
        Returns the application command's guild.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    def iter_options(self):
        """
        Iterates over the options of the application command.
        
        This method is an iterable generator.
        
        Yields
        ------
        option : ``ApplicationCommandOption``
        """
        options = self.options
        if (options is not None):
            yield from options



from .attachment import *
from .message import *
from .message_activity import *
from .message_application import *
from .message_call import *
from .message_interaction import *
from .message_role_subscription import *


__all__ = (
    *attachment.__all__,
    *message.__all__,
    *message_activity.__all__,
    *message_application.__all__,
    *message_call.__all__,
    *message_interaction.__all__,
    *message_role_subscription.__all__,
)



__all__ = ()

from ...channel import (
    ChannelFlag, ChannelType, ForumLayout, ForumTag, PermissionOverwrite, SortOrder, VideoQualityMode, VoiceRegion
)
from ...emoji import create_emoji_from_exclusive_data

from ..audit_log_change import AuditLogChange

from .shared import _convert_preinstanced, convert_nothing, convert_snowflake, convert_snowflake_array


def convert_snowflake_array__applied_tag_ids(name, data):
    return convert_snowflake_array('applied_tag_ids', data)


def convert_bool__open(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = not before
    
    after = data.get('new_value', None)
    if (after is not None):
        after = not after
    
    return AuditLogChange('open', before, after)


def convert_channel_flags(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = ChannelFlag(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = ChannelFlag(after)
    
    return AuditLogChange('flags', before, after)


def convert_int__auto_archive_after(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before *= 60
    
    after = data.get('new_value', None)
    if (after is not None):
        after *= 60
    
    return AuditLogChange('auto_archive_after', before, after)


def convert_forum_tags(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        if before:
            before = (*sorted(ForumTag.from_data(tag_data) for tag_data in before),)
        else:
            before = None
    
    after = data.get('new_value', None)
    if (after is not None):
        if after:
            after = (*sorted(ForumTag.from_data(tag_data) for tag_data in after),)
        else:
            after = None
    
    return AuditLogChange('available_tags', before, after)


def convert_int__default_thread_auto_archive_after(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before *= 60
    
    after = data.get('new_value', None)
    if (after is not None):
        after *= 60
    
    return AuditLogChange('default_thread_auto_archive_after', before, after)


def convert_int__slowmode(name, data):
    return convert_nothing('slowmode', data)


def convert_int__default_thread_slowmode(name, data):
    return convert_nothing('default_thread_slowmode', data)


def convert_permission_overwrites(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        if before:
            before = {
                permission_overwrite.id: permission_overwrite for permission_overwrite in
                (PermissionOverwrite.from_data(overwrite_data) for overwrite_data in before)
            }
        else:
            before = None
    
    after = data.get('new_value', None)
    if (after is not None):
        if after:
            after = {
                permission_overwrite.id: permission_overwrite for permission_overwrite in
                (PermissionOverwrite.from_data(overwrite_data) for overwrite_data in after)
            }
        else:
            after = None
    
    return AuditLogChange('permission_overwrites', before, after)


def convert_video_quality_mode(name, data):
    return _convert_preinstanced('video_quality_mode', data, VideoQualityMode)


def convert_voice_region(name, data):
    return _convert_preinstanced('region', data, VoiceRegion)


def convert_default_thread_reaction(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = create_emoji_from_exclusive_data(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = create_emoji_from_exclusive_data(after)
    
    return AuditLogChange('default_thread_reaction', before, after) 


def default_sort_order(name, data):
    return _convert_preinstanced('default_sort_order', data, SortOrder)


def default_forum_layout(name, data):
    return _convert_preinstanced('default_forum_layout', data, ForumLayout)


def convert_channel_type(name, data):
    return _convert_preinstanced('type', data, ChannelType)


CHANNEL_CONVERTERS = {
    'applied_tags': convert_snowflake_array__applied_tag_ids,
    'archived': convert_nothing,
    'auto_archive_duration': convert_int__auto_archive_after,
    'available_tags': convert_forum_tags,
    'bitrate': convert_nothing,
    'default_forum_layout': default_forum_layout,
    'default_sort_order': default_sort_order,
    'default_auto_archive_duration': convert_int__default_thread_auto_archive_after,
    'default_reaction_emoji': convert_default_thread_reaction,
    'default_thread_rate_limit_per_user': convert_int__default_thread_slowmode,
    'flags': convert_channel_flags,
    'invitable': convert_nothing,
    'locked': convert_bool__open,
    'name': convert_nothing,
    'nsfw': convert_nothing,
    'rate_limit_per_user': convert_int__slowmode,
    'parent_id': convert_snowflake,
    'permission_overwrites': convert_permission_overwrites,
    'position': convert_nothing,
    'rtc_region': convert_voice_region,
    'topic': convert_nothing,
    'type': convert_channel_type,
    'video_quality_mode': convert_video_quality_mode,
    'user_limit': convert_nothing,
}



from .fields import *
from .flags import *
from .thread_profile import *
from .utils import *


__all__ = (
    *fields.__all__,
    *flags.__all__,
    *thread_profile.__all__,
    *utils.__all__,
)



__all__ = ('cr_p_permission_overwrite_object', )

import warnings

from scarletio import include

from .constants import PERMISSION_ALLOW_KEY, PERMISSION_DENY_KEY

PermissionOverwriteTargetType = include('PermissionOverwriteTargetType')
Role = include('Role')


def cr_p_permission_overwrite_object(target, allow, deny):
    """
    Creates a json serializable object representing a ``PermissionOverwrite``.
    
    Deprecated.
    
    Parameters
    ----------
    target : ``ClientUserBase``, ``Role``
        The target entity of the overwrite.
        The allowed permissions by the overwrite.
    deny : `int`
        The denied permission by the overwrite.
    
    Returns
    -------
    permission_overwrite_data : `dict` of (`str`, `Any) items
    """
    warnings.warn(
        (
            f'`{cr_p_permission_overwrite_object}` is deprecated and will be removed in 2023 April. '
            f'Please use `PermissionOverwrite(...).to_data(...)` instead.'
        ),
        FutureWarning,
        stacklevel = 2,
    )

    if isinstance(target, Role):
        permission_overwrite_target_type = PermissionOverwriteTargetType.role
    else:
        permission_overwrite_target_type = PermissionOverwriteTargetType.user
    
    return {
        PERMISSION_ALLOW_KEY: allow,
        PERMISSION_DENY_KEY: deny,
        'id': target.id,
        'type': permission_overwrite_target_type.value,
    }



__all__ = ('EmbedImage',)

from scarletio import copy_docs

from ...utils import url_cutter

from ..embed_field_base import EmbedFieldBase

from .fields import (
    parse_height, parse_proxy_url, parse_url, parse_width, put_height_into, put_proxy_url_into, put_url_into,
    put_width_into, validate_url
)


class EmbedImage(EmbedFieldBase):
    """
    Represents an embed's image.
    
    Attributes
    ----------
    height : `int`
        The height of the image. Defaults to `0`.
    proxy_url : `None`, `str`
        A proxied url of the image.
    url : `None`, `str`
        The url of the image.
    width : `int`
        The width of the image. Defaults to `0`.
    """
    __slots__ = ('height', 'proxy_url', 'url', 'width',)
    
    def __new__(cls, url = ...):
        """
        Creates a new embed field with the given parameters.
        
        Parameters
        ----------
        url : `None`, `str`, Optional
            The url of the image. Can be http(s) or attachment.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # url
        if url is ...:
            url = None
        else:
            url = validate_url(url)
        
        self = object.__new__(cls)
        self.height = 0
        self.proxy_url = None
        self.url = url
        self.width = 0
        return self
    
    
    @copy_docs(EmbedFieldBase.__bool__)
    def __bool__(self):
        # url
        if self.url is not None:
            return True
        
        return False
    
    
    @copy_docs(EmbedFieldBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts.append(' url = ')
        url = self.url
        if url is None:
            repr_parts.append('null')
        else:
            repr_parts.append(repr(url_cutter(url)))
        
        width = self.width
        height = self.height
        if width or height:
            repr_parts.append(', size = ')
            repr_parts.append(str(self.width))
            repr_parts.append('x')
            repr_parts.append(str(self.height))
    
    
    @copy_docs(EmbedFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # url
        url = self.url
        if (url is not None):
            hash_value ^= hash(url)
        
        return hash_value
    
    
    @copy_docs(EmbedFieldBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # height
        # Receive only | check `.url` instead
        
        # proxy_url
        # Receive only | check `.url` instead
        
        # url
        if self.url != other.url:
            return False
        
        # width
        # Receive only | check `.url` instead
        
        return True
    
    
    @classmethod
    @copy_docs(EmbedFieldBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.height = parse_height(data)
        self.url = parse_url(data)
        self.proxy_url = parse_proxy_url(data)
        self.width = parse_width(data)
        return self
    
    
    @copy_docs(EmbedFieldBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = {}
        
        put_url_into(self.url, data, defaults)
        
        if include_internals:
            put_height_into(self.height, data, defaults)
            put_proxy_url_into(self.proxy_url, data, defaults)
            put_width_into(self.width, data, defaults)
        
        return data
    
    
    @copy_docs(EmbedFieldBase.clean_copy)
    def clean_copy(self, guild = None):
        new = object.__new__(type(self))
        new.height = self.height
        new.proxy_url = self.proxy_url
        new.url = self.url
        new.width = self.width
        return new
    
    
    @copy_docs(EmbedFieldBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.height = self.height
        new.proxy_url = self.proxy_url
        new.url = self.url
        new.width = self.width
        return new
    
    
    def copy_with(self, *, url = ...):
        """
        Copies the embed image with the given parameters.
        
        Parameters
        ----------
        url : `None`, `str`, Optional (Keyword only)
            The url of the image. Can be http(s) or attachment.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # url
        if url is ...:
            url = self.url
            height = self.height
            proxy_url = self.proxy_url
            width = self.width
        else:
            url = validate_url(url)
            height = 0
            proxy_url = None
            width = 0
        
        new = object.__new__(type(self))
        new.height = height
        new.proxy_url = proxy_url
        new.url = url
        new.width = width
        return new



__all__ = ()

from ...emoji import Emoji, create_partial_emoji_data, create_partial_emoji_from_data
from ...field_parsers import entity_id_parser_factory, nullable_functional_parser_factory, preinstanced_parser_factory
from ...field_putters import (
    entity_id_putter_factory, nullable_functional_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, nullable_entity_validator_factory, preinstanced_validator_factory
)
from ...user import ClientUserBase

from ..channel import Channel

from .preinstanced import VoiceChannelEffectAnimationType

# animation_id

parse_animation_id = entity_id_parser_factory('animation_id')
put_animation_id_into = entity_id_putter_factory('animation_id')
validate_animation_id = entity_id_validator_factory('animation_id')

# animation_type

parse_animation_type = preinstanced_parser_factory(
    'animation_type', VoiceChannelEffectAnimationType, VoiceChannelEffectAnimationType.premium
)
put_animation_type_into = preinstanced_putter_factory('animation_type')
validate_animation_type = preinstanced_validator_factory('animation_type', VoiceChannelEffectAnimationType)

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id_into = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)

# emoji

parse_emoji = nullable_functional_parser_factory('emoji', create_partial_emoji_from_data)
put_emoji_into = nullable_functional_optional_putter_factory('emoji', create_partial_emoji_data)
validate_emoji = nullable_entity_validator_factory('emoji', Emoji)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id_into = entity_id_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)



__all__ = ('ApplicationCommandOptionMetadataSubCommand',)

from scarletio import copy_docs

from .fields import parse_default, put_default_into, validate_default
from .nested import ApplicationCommandOptionMetadataNested


class ApplicationCommandOptionMetadataSubCommand(ApplicationCommandOptionMetadataNested):
    """
    Sub-command parameter application command option metadata.
    
    Attributes
    ----------
    default : `bool`
        Whether the option is the default one.
    
    options : `None`, `tuple` of ``ApplicationCommandOption``
        Contains the option's parameter or its sub-commands (or groups).
    """
    __slots__ = ('default',)
    
    
    def __new__(cls, *, default = ..., options = ...):
        """
        Creates a new sub-command application command option metadata with the given parameters.
        
        Parameters
        ----------
        default : `bool`, Optional (Keyword only)
            Whether the option is the default one.
        
        options : `None`, `iterable` of ``ApplicationCommandOption``, Optional (Keyword only)
            Contains the option's parameter or its sub-commands (or groups).
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # default
        if default is ...:
            default = False
        else:
            default = validate_default(default)
        
        # Construct
        new = ApplicationCommandOptionMetadataNested.__new__(
            cls, 
            options = options,
        )
        new.default = default
        return new
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataNested.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            default = keyword_parameters.pop('default', ...),
            options = keyword_parameters.pop('options', ...),
        )
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataNested.from_data)
    def from_data(cls, data):
        self = super(ApplicationCommandOptionMetadataSubCommand, cls).from_data(data)
        self.default = parse_default(data)
        return self
    
    
    @copy_docs(ApplicationCommandOptionMetadataNested.to_data)
    def to_data(self, *, defaults = False):
        data = ApplicationCommandOptionMetadataNested.to_data(self, defaults = defaults)
        put_default_into(self.default, data, defaults)
        return data
    
    
    @copy_docs(ApplicationCommandOptionMetadataNested._add_type_specific_repr_fields)
    def _add_type_specific_repr_fields(self, repr_parts):
        ApplicationCommandOptionMetadataNested._add_type_specific_repr_fields(self, repr_parts)
        
        repr_parts.append(', default = ')
        repr_parts.append(repr(self.default))
    
    
    @copy_docs(ApplicationCommandOptionMetadataNested._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ApplicationCommandOptionMetadataNested._is_equal_same_type(self, other):
            return False
        
        if self.default != other.default:
            return False
        
        return True
    
    
    @copy_docs(ApplicationCommandOptionMetadataNested.__hash__)
    def __hash__(self):
        hash_value = ApplicationCommandOptionMetadataNested.__hash__(self)
        
        # default
        hash_value ^= self.default << 13
        
        return hash_value
    
    
    @copy_docs(ApplicationCommandOptionMetadataNested.copy)
    def copy(self):
        new = ApplicationCommandOptionMetadataNested.copy(self)
        new.default = self.default
        return new
    
    
    def copy_with(self, *, default = ..., options = ...):
        """
        Copies the sub-command application command option metadata with the given fields.
        
        Parameters
        ----------
        default : `bool`, Optional (Keyword only)
            Whether the option is the default one.
        
        options : `None`, `iterable` of ``ApplicationCommandOption``, Optional (Keyword only)
            Contains the option's parameter or its sub-commands (or groups).
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # default
        if default is ...:
            default = self.default
        else:
            default = validate_default(default)
        
        # Construct
        new = ApplicationCommandOptionMetadataNested.copy_with(
            self,
            options = options,
        )
        new.default = default
        return new
    
    
    @copy_docs(ApplicationCommandOptionMetadataNested.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            default = keyword_parameters.pop('default', ...),
            options = keyword_parameters.pop('options', ...),
        )



__all__ = ()

import sys
from platform import platform as get_platform

from scarletio import __package__ as SCARLETIO_NAME, __version__ as SCARLETIO_VERSION

from .... import __package__ as PACKAGE_NAME, __version__ as PACKAGE_VERSION

from ...core import register


@register(
    alters = 'v'
)
def version():
    """Prints out hata\'s version."""
    output_parts = []
    
    output_parts.append(PACKAGE_NAME[0].upper())
    output_parts.append(PACKAGE_NAME[1:])
    output_parts.append(' version: ')
    output_parts.append(PACKAGE_VERSION)
    output_parts.append('\n')
    
    output_parts.append(SCARLETIO_NAME[0].upper())
    output_parts.append(SCARLETIO_NAME[1:])
    output_parts.append(' version: ')
    output_parts.append(SCARLETIO_VERSION)
    output_parts.append('\n')
    
    output_parts.append('Platform: ')
    output_parts.append(get_platform())
    output_parts.append('\n')
    
    implementation = sys.implementation
    
    output_parts.append('Python implementation: ')
    output_parts.append(implementation.name)
    output_parts.append('\n')
    
    version = implementation.version
    
    output_parts.append('Python version: ')
    output_parts.append(str(version[0]))
    output_parts.append('.')
    output_parts.append(str(version[1]))
    
    if version[3] == 'final':
        output_parts.append(' final')
    
    output_parts.append('\n')
    
    return ''.join(output_parts)



from .connection import *
from .oauth2_access import *
from .oauth2_user import *

from .achievement import *
from .helpers import *


__all__ = (
    *connection.__all__,
    *oauth2_access.__all__,
    *oauth2_user.__all__,
    
    *achievement.__all__,
    *helpers.__all__,
)

from ...utils.module_deprecation import deprecated_import
deprecated_import(Oauth2User, 'UserOA2')
deprecated_import(Oauth2Access, 'OA2Access')



__all__ = ('ApplicationCommandOptionMetadataChannel',)

from scarletio import copy_docs

from .fields import parse_channel_types, put_channel_types_into, validate_channel_types
from .parameter import ApplicationCommandOptionMetadataParameter


class ApplicationCommandOptionMetadataChannel(ApplicationCommandOptionMetadataParameter):
    """
    Channel parameter application command option metadata.
    
    Parameters
    ----------
    channel_types : `None`, `tuple` of ``ChannelType``
        The accepted channel types by the option. Defaults to `None`
    
    required : `bool`
        Whether the parameter is required. Defaults to `False`.
    """
    __slots__ = ('channel_types',)
    
    def __new__(cls, *, channel_types = ..., required = ...):
        """
        Creates a new channel application command option metadata with the given parameters.
        
        Parameters
        ----------
        channel_types : `None`, `iterable` of (``ChannelType``, `int`), Optional (Keyword only)
            The accepted channel types by the option.
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # channel_types
        if channel_types is ...:
            channel_types = None
        else:
            channel_types = validate_channel_types(channel_types)
        
        # Construct
        new = ApplicationCommandOptionMetadataParameter.__new__(
            cls,
            required = required,
        )
        new.channel_types = channel_types
        return new
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataParameter.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            channel_types = keyword_parameters.pop('channel_types', ...),
            required = keyword_parameters.pop('required', ...),
        )
    
    
    @classmethod
    @copy_docs(ApplicationCommandOptionMetadataParameter.from_data)
    def from_data(cls, data):
        self = super(ApplicationCommandOptionMetadataChannel, cls).from_data(data)
        self.channel_types = parse_channel_types(data)
        return self
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter.to_data)
    def to_data(self, *, defaults = False):
        data = ApplicationCommandOptionMetadataParameter.to_data(self, defaults = defaults)
        put_channel_types_into(self.channel_types, data, defaults)
        return data
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter._add_type_specific_repr_fields)
    def _add_type_specific_repr_fields(self, repr_parts):
        ApplicationCommandOptionMetadataParameter._add_type_specific_repr_fields(self, repr_parts)
        
        repr_parts.append(', channel_types = ')
        repr_parts.append(repr(self.channel_types))
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ApplicationCommandOptionMetadataParameter._is_equal_same_type(self, other):
            return False
        
        if self.channel_types != other.channel_types:
            return False
        
        return True
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter.__hash__)
    def __hash__(self):
        hash_value = ApplicationCommandOptionMetadataParameter.__hash__(self)
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            hash_value ^= len(channel_types) << 12
            
            for channel_type in channel_types:
                hash_value ^= hash(channel_type)
        
        return hash_value
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter.copy)
    def copy(self):
        new = ApplicationCommandOptionMetadataParameter.copy(self)
        
        channel_types = self.channel_types
        if (channel_types is not None):
            channel_types = (*channel_types,)
        new.channel_types = channel_types
        
        return new
    
    
    def copy_with(self,  *, channel_types = ..., required = ...):
        """
        Copies the channel application command option metadata with the given fields.
        
        Parameters
        ----------
        channel_types : `None`, `iterable` of (``ChannelType``, `int`), Optional (Keyword only)
            The accepted channel types by the option.
        
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter of incorrect type given.
        ValueError
            - If a parameter of incorrect value given.
        """
        # channel_types
        if channel_types is ...:
            channel_types = self.channel_types
            if (channel_types is not None):
                channel_types = (*channel_types,)
        else:
            channel_types = validate_channel_types(channel_types)
        
        # Construct
        new = ApplicationCommandOptionMetadataParameter.copy_with(
            self,
            required = required,
        )
        new.channel_types = channel_types
        return new
    
    
    @copy_docs(ApplicationCommandOptionMetadataParameter.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            channel_types = keyword_parameters.pop('channel_types', ...),
            required = keyword_parameters.pop('required', ...),
        )
